/* -------------------------------------------------------------------------------------------------------------------------
* Project           :C-to-FPGA Solution for Acclerating Tomographic Reconstruction
* File Name         :uwer_forward.c (from project UW Emission Reconstruction) 
*                    Copyright (c) 2004 University of Washington - Imaging Research Laboratory. All rights reserved.
* Version           :1.0 (V1)
* (V1)Author        :Adam Alessio
* (V1)Created       :Wed, Jan 14 2004
* References        :
* Description       :Performs forward projection
*
******This file may not be distributed or modified without this header*****
*-------------------------------------------------------------------------------------------------------------------------*/


#include "uwer_forward.h"

void forwardSimulation(struct sino_info *sinoInfo,
		       struct recon_image_info *imageInfo,
		       struct exec_parameters *params)
{
  int frame, plane;
  float **sinogramData, **image;
  
  Print("-----Performing simple forward projection\n");


   for (frame=0;frame<imageInfo->nFrames;frame++)
    for (plane=0;plane<imageInfo->zDim;plane++)
      {
	printf("Forward projecting frame:%d of %d and plane:%d of %d\n", frame, imageInfo->nFrames,plane,imageInfo->zDim);

	sinogramData = sinoInfo->sinogramData[frame][plane];
	image = imageInfo->imageData[frame][plane];

	
	forwardproject(image,sinogramData,sinoInfo,imageInfo);
      }
   output_sinogram(params->InputSinogramFileName,params, sinoInfo);
   fflush(stdout);fflush(stderr);
}

//Uncomment one of the 3 versions of this function
// 1. Original
// 2. Uses both ray equations for all angles
// 3. Swaps ray equations based on which angle gives better projections


/* Forward projector designed to match simple back-projector used in uwer_fbp.c */
/* This projector does not support projections with a kernel shape as in uwer_SystemMatrix.c */
// 1. Original function
void forwardproject(float **image,float **sinogramData,struct sino_info *sino,struct recon_image_info *geom)
{
  double x=0.0,y=0.0;
  double theta,t,dat,findex;
  int i,j,k,index;
  double *cs,*sn; 
       /* arrays to precompute cos(theta_k), sin(theta_k) */
  double t_0;
       /* initial projection displacement */
       /* t_0 = - (nProj-1)/2*projSpacing;*/
  float *im_pt;
  float val,val2; 
  /* pointer to imageio[i][j]  Assume data is */
  /* allocated as 1-D array w/ pointers on each row */
  
  
   /* Precompute sin and cos for all projection angles */
  cs=(double*)malloc(sizeof(double)*(double)sino->nAngles);
  sn=(double*)malloc(sizeof(double)*(double)sino->nAngles);
  theta=-180.0*(double)M_PI/360.0;  // Rotate phantom 90 degrees
  for (k=0;k<sino->nAngles;k++){
     sn[k]=sin(theta);
     cs[k]=cos(theta);
     theta+=(sino->angleSpacing);   /* AMA Remove conversion, already in rads *(double)M_PI/180.0); */
  }

  /* Start Forward Projection */

  //  printf("Image Value %f\n",image[256][256]);
  
  im_pt=&image[0][0];

  //  printf("Image Value %f\n",*im_pt);

  // IS THIS T_0 ACCURATE?
  // Depends on data leave odd versus even setting for flexibility  - Both even and odd identical for now
  if(sino->nProjs & 1)
    t_0 = -0.5*(double)(sino->nProjs-1)*sino->projSpacing;
  else
    t_0 = (double)((-0.5*(sino->nProjs-1)))*sino->projSpacing;
 
    //    t_0 = (double)((int)(-0.5*(sino->nProjs-1)))*sino->projSpacing-sino->projSpacing;
    // t_0 = (double)((int)(-0.5*(sino->nProjs)))*sino->projSpacing;  // Same as above
 
  printf("projSpacing = %f, nProjs = %d\n",sino->projSpacing,sino->nProjs);
  printf("Pixel Size: %f,%f\n",  geom->yPixelSize,geom->xPixelSize);
  y=t_0;
  
  for (i=0;i<geom->yDim;i++) // loop through y's
    {
      x=t_0;
      for (j=0;j<geom->xDim;j++)  // loop through x's
	{
	  for (k=0;k<sino->nAngles;k++)
	    {
	      t=y*cs[k]-x*sn[k];    
	      findex=(t-t_0)/sino->projSpacing;
	      /* Check whether location is out of */
              /* projection area for this angle */
	      /* If so,  set image[i][j] to zero, */
              /* break k loop and go to next image location */
	      if ((findex>=0) && (int)findex < (sino->nProjs - 2) )
		{
		  // Interpolate pixel between two closest rays
		  index=(int)findex;
		  // Pixelization errors in sinogram are there with and without interpolation
		  if ((findex-(double)index > -1E-10) && (findex-(double)index<1E-10))
		    sinogramData[k][(int)findex] += (*im_pt);
		  else /* if not integer, then bilinear interpolation between parallel projections */
		    {
		      val = ((double)((int)(findex+1.0)))-findex;
		      val2 = 1.0-val;
		  
		      sinogramData[k][(int)findex] += val* (*im_pt);
		      sinogramData[k][(int)findex+1] += val2* (*im_pt);
		    }
		}	      
	    }
	  
	  im_pt++;
	  x += geom->xPixelSize;
	}
      y += geom->yPixelSize;
    }
    
  free((void*)cs);
  free((void*)sn);
}


// //2.Forward Projector that uses both ray equations for all angles

// void forwardproject(float **image,float **sinogramData,struct sino_info *sino,struct recon_image_info *geom)
// {
//   double x=0.0,y=0.0;
//   double theta,t,dat,findex,yindex,xindex;
//   int i,j,k,index,s;
//   double *cs,*sn,*tn,*ce,*sc,*ct; 

//        /* arrays to precompute cos(theta_k), sin(theta_k) */
//   double t_0;
//        /* initial projection displacement */
//        /* t_0 = - (nProj-1)/2*projSpacing;*/

//    int swap_check; 
//   float val,val2; 
//   /* pointer to imageio[i][j]  Assume data is */
//   /* allocated as 1-D array w/ pointers on each row */
  
  
//    /* Precompute sin and cos for all projection angles */
//   cs=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   sn=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   tn=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   ce=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   sc=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   ct=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   //theta=-180.0*(double)M_PI/360.0;  // Rotate phantom 90 degrees
//   theta=(double)M_PI;  // Rotate phantom 180 degrees
//   //theta=0;
  
//   for (k=0;k<sino->nAngles;k++){
//      sn[k]=sin(theta);
//      cs[k]=cos(theta);
//      tn[k]=sn[k]/cs[k];
//      ce[k]=1/sn[k];
//      sc[k]=1/cs[k];
//      ct[k]=cs[k]/sn[k];
//      theta+=(sino->angleSpacing);   /* AMA Remove conversion, already in rads *(double)M_PI/180.0); */
//      printf("\ntheta[%d]=%f",k,theta);
//   }

//   /* Start Forward Projection */

//   printf("\nAngle spaceing = %f",sino->angleSpacing);
//   // IS THIS T_0 ACCURATE?
//   // Depends on data leave odd versus even setting for flexibility  - Both even and odd identical for now
//   if(sino->nProjs & 1)
//     t_0 = -0.5*(double)(sino->nProjs-1)*sino->projSpacing;
//   else
//     t_0 = (double)((-0.5*(sino->nProjs-1)))*sino->projSpacing;
 
//   printf("projSpacing = %f, nProjs = %d\n",sino->projSpacing,sino->nProjs);
//   printf("\n Image Dim = %dx%d",geom->xDim,geom->yDim);
//   printf("Pixel Size: %f,%f\n",  geom->yPixelSize,geom->xPixelSize);
//   printf("\nt_0: %f\n",  t_0);
//   printf("AngleSpacing = %f, Number Angles = %d\n",sino->angleSpacing,sino->nAngles);

  
//   for (k=0;k<sino->nAngles;k++)
//     for (s=0;s<sino->nProjs;s++)  
//     {  sinogramData[k][s]=0;

//     }

//   for (k=0;k<sino->nAngles;k++)
//     { 
//        for (s=0;s<sino->nProjs;s++)  
// 	    { t = (s*sino->projSpacing) + t_0;  
//           y=t_0;
//           for (i=0;i<geom->yDim;i++) 
// 	        { 
// 	          x = y*ct[k]-t*ce[k]; 
// 	          xindex = (x-t_0)/geom->xPixelSize;;
//            index= (int)xindex;
// 	          /* Check whether location is out of */
//            /* projection area for this angle */
// 	          /* If so,  set image[i][j] to zero, */
//            /* break k loop and go to next image location */
// 	          if ((index>=0) && (index < (geom->xDim- 1)) )
// 		        {  
// 		          // Interpolate pixel between two closest rays
// 		          // Pixelization errors in sinogram are there with and without interpolation
// 		          val = ((double)(index+1))-xindex;
// 		          val2 = 1.0-val;
//                   sinogramData[k][s] += val*image[index][i] + val2*image[index+1][i];
// 		        }
//               y+=geom->yPixelSize; 	      
// 	        }
//         } 
//        for (s=0;s<sino->nProjs;s++)  
//  	   { 
//           t = (s*sino->projSpacing) + t_0;   
//   		  x=t_0;
//           for (i=0;i<geom->xDim;i++) 
// 	        { 
// 	          y = x*tn[k]+t*sc[k];    
// 	          yindex = (y-t_0)/geom->yPixelSize;;
//               index= (int)yindex;
    
// 	          /* Check whether location is out of */
//            /* projection area for this angle */
// 	          /* If so,  set image[i][j] to zero, */
//            /* break k loop and go to next image location */
// 	          if ((index>=0) && (index < (geom->yDim- 1)) )
// 		        {  
// 		          // Interpolate pixel between two closest rays
// 		          // Pixelization errors in sinogram are there with and without interpolation
// 		          val = ((double)(index+1))-yindex;
// 		          val2 = 1.0-val;
//                sinogramData[k][s] += val*image[i][index] + val2*image[i][index+1];
// 		        }
//               x+=geom->xPixelSize; 	      
// 	        }
//        }
//     }   
  
//   free((void*)cs);
//   free((void*)sn);
//   free((void*)tn);
//   free((void*)ce);
//   free((void*)sc);
//   free((void*)ct);
// }

// //3.Forward Projector that uses 2 different ray equations for 45 degree sections of the image

// void forwardproject(float **image,float **sinogramData,struct sino_info *sino,struct recon_image_info *geom)
// {
//   double x=0.0,y=0.0;
//   double theta,t,dat,findex,yindex,xindex;
//   int i,j,k,index,s;
//   double *cs,*sn,*tn,*ce,*sc,*ct; 
//        /* arrays to precompute cos(theta_k), sin(theta_k) */
//   double t_0;
//        /* initial projection displacement */
//        /* t_0 = - (nProj-1)/2*projSpacing;*/
//   int swap_check; 
//   float val,val2; 
//   /* pointer to imageio[i][j]  Assume data is */
//   /* allocated as 1-D array w/ pointers on each row */
  
  
//    /* Precompute sin and cos for all projection angles */
//   cs=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   sn=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   tn=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   ce=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   sc=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   ct=(double*)malloc(sizeof(double)*(double)sino->nAngles);
//   //theta=-180.0*(double)M_PI/360.0;  // Rotate phantom 90 degrees
//   theta=(double)M_PI;  // Rotate phantom 180 degrees
//   //theta=0;
  
//   for (k=0;k<sino->nAngles;k++){
//      sn[k]=sin(theta);
//      cs[k]=cos(theta);
//      tn[k]=sn[k]/cs[k];
//      ce[k]=1/sn[k];
//      sc[k]=1/cs[k];
//      ct[k]=cs[k]/sn[k];
//      theta+=(sino->angleSpacing);   /* AMA Remove conversion, already in rads *(double)M_PI/180.0); */
//      printf("\ntheta[%d]=%f",k,theta);
//   }

//   /* Start Forward Projection */

//   printf("\nAngle spaceing = %f",sino->angleSpacing);
//   // IS THIS T_0 ACCURATE?
//   // Depends on data leave odd versus even setting for flexibility  - Both even and odd identical for now
//   if(sino->nProjs & 1)
//     t_0 = -0.5*(double)(sino->nProjs-1)*sino->projSpacing;
//   else
//     t_0 = (double)((-0.5*(sino->nProjs-1)))*sino->projSpacing;
 
//   printf("projSpacing = %f, nProjs = %d\n",sino->projSpacing,sino->nProjs);
//   printf("\n Image Dim = %dx%d",geom->xDim,geom->yDim);
//   printf("Pixel Size: %f,%f\n",  geom->yPixelSize,geom->xPixelSize);
//   printf("\nt_0: %f\n",  t_0);
//   printf("AngleSpacing = %f, Number Angles = %d\n",sino->angleSpacing,sino->nAngles);

  
//   for (k=0;k<sino->nAngles;k++)
//     for (s=0;s<sino->nProjs;s++)  
//       sinogramData[k][s]=0;


//   swap_check=sino->nAngles/4;

//   for (k=0;k<sino->nAngles;k++)
//     { 
     
//       for (s=0;s<sino->nProjs;s++)  
// 	  { 
//         t = (s*sino->projSpacing) + t_0;  
        
//         if(((k>=swap_check)&&(k<(3*swap_check))))
//         { 
//           y=t_0;
//           for (i=0;i<geom->yDim;i++) 
// 	        { 
// 	          x = y*ct[k]-t*ce[k]; 
// 	          xindex = (x-t_0)/geom->xPixelSize;;
//               index= (int)xindex;
//               //printf("\nimage[%d][%d]",i,index);
// 	          /* Check whether location is out of */
//               /* projection area for this angle */
// 	          /* If so,  set image[i][j] to zero, */
//               /* break k loop and go to next image location */
// 	          if ((index>=0) && (index < (geom->xDim- 1)) )
// 		        {  
// 		          // Interpolate pixel between two closest rays
// 		          // Pixelization errors in sinogram are there with and without interpolation
// 		          val = ((double)(index+1))-xindex;
// 		          val2 = 1.0-val;
//                   //printf("\nk = %d, s = %d,image [%d][%d] = %f, val = %f, val2 = %f",k,s,i,index,image[i][index], val, val2);
//                   sinogramData[k][s] += val*image[index][i] + val2*image[index+1][i];
//                   //printf("\nk = %d, s = %d,image[%d][%d], sino[%d][%d] = %f",k,s,i,index,k,s,sinogramData[k][s]);
// 		        }
//               y+=geom->xPixelSize; 	      
// 	        }
//         }
//         else //if(((k<swap_check)&&(k>(3*swap_check))))
//         { 
// 		  x=t_0;
//           for (i=0;i<geom->xDim;i++) 
// 	        { 
// 	          y = x*tn[k]+t*sc[k];    
// 	          yindex = (y-t_0)/geom->yPixelSize;;
//               index= (int)yindex;
//               //printf("\nimage[%d][%d]",i,index);
// 	          /* Check whether location is out of */
//               /* projection area for this angle */
// 	          /* If so,  set image[i][j] to zero, */
//               /* break k loop and go to next image location */
// 	          if ((index>=0) && (index < (geom->xDim- 1)) )
// 		        {  
// 		          // Interpolate pixel between two closest rays
// 		          // Pixelization errors in sinogram are there with and without interpolation
// 		          val = ((double)(index+1))-yindex;
// 		          val2 = 1.0-val;
//                   //printf("\nk = %d, s = %d,image [%d][%d] = %f, val = %f, val2 = %f",k,s,i,index,image[i][index], val, val2);
//                   sinogramData[k][s] += val*image[i][index] + val2*image[i][index+1];
//                   //printf("\nk = %d, s = %d,image[%d][%d], sino[%d][%d] = %f",k,s,i,index,k,s,sinogramData[k][s]);
// 		        }
//               x+=geom->xPixelSize; 	      
// 	        }
//         }
//       }   
// 	}
//   free((void*)cs);
//   free((void*)sn);
//   free((void*)tn);
//   free((void*)ce);
//   free((void*)sc);
//   free((void*)ct);
// }

